@@ -4,20 +4,15 @@ import os
 import sys
 import getpass
 from six import reraise
-from six.moves import configparser
 from io import StringIO
 from functools import wraps
-from sprinter.core import PHASE
-import sprinter.brew as brew
+from collections import defaultdict
+
 import sprinter.lib as lib
-from sprinter.formulabase import FormulaBase
-from sprinter.directory import Directory
-from sprinter.exceptions import SprinterException
-from sprinter.injections import Injections
-from sprinter.manifest import Manifest
-from sprinter.system import System
-from sprinter.pippuppet import Pip, PipException
-from sprinter.templates import shell_utils_template, source_template, warning_template
+from sprinter.core import PHASE, load_global_config, Directory, Injections, Manifest, load_manifest, FeatureDict
+from sprinter.core.templates import shell_utils_template, source_template
+from sprinter.lib import SprinterException, system
+from sprinter.external import brew
 
 
 def warmup(f):
@@ -82,49 +77,48 @@ class Environment(object):
     directory = None  # handles interactions with the environment directory
     injections = None  # handles injections
     global_injections = None  # handles injections for the global sprinter configuration
-    system = None  # stores utility methods to determine system specifics
     # variables typically populated programatically
     warmed_up = False  # returns true if the environment is ready for environments
     shell_util_path = None  # the path to the shell utils file
     error_occured = False
-    # A dictionary for class object instances. Exists Mainly for testability + injection
-    formula_dict = {}
-    # a dictionary of the feature objects.
-    # The key is a tuple of feature name and formula, while the value is an instance.
-    _feature_dict = {}
-    # the order of the feature dict.
-    _feature_dict_order = []
-    # a dictionary of the errors associated with features.
-    # The key is a tuple of feature name and formula, while the value is an instance.
-    _error_dict = {}
     _errors = []  # list to keep all the errors
-    # a pip puppet used to install eggs and add it to the classpath
-    _pip = None
     sandboxes = []  # a list of package managers to sandbox (brew)
     # specifies where to get the global sprinter root
     global_config = None  # configuration file, which defaults to loading from SPRINTER_ROOT/.global/config.cfg
-    write_files = True  # write files to the filesystem.
     ignore_errors = False  # ignore errors in features
 
-    def __init__(self, logger=None, logging_level=logging.INFO,
-                 root=None, sprinter_namespace='sprinter',
-                 global_config=None, write_files=True,
+    def __init__(self,
+                 logger=None,
+                 logging_level=logging.INFO,
+                 root=None,
+                 sprinter_namespace=None,
+                 global_config=None,
                  ignore_errors=False):
-        self.system = System()
-        if not logger:
-            logger = self._build_logger(level=logging_level)
-        self.logger = logger
-        self.sprinter_namespace = sprinter_namespace
-        self.root = root or os.path.expanduser(os.path.join("~", ".%s" % sprinter_namespace))
-        self.global_path = os.path.join(self.root, ".global")
-        self._pip = Pip(self.global_path)
+
+        # base logging object to log instances
+        self.logger = logger or self._build_logger(level=logging_level)
         if logging_level == logging.DEBUG:
             self.logger.info("Starting in debug mode...")
-        self.formula_dict = {}
-        self.shell_util_path = os.path.join(self.global_path, "utils.sh")
-        self.load_global_config(global_config)
-        self.write_files = write_files
+
+        # the sprinter namespace
+        self.sprinter_namespace = sprinter_namespace or 'sprinter'
+
+        # the root directory which sprinter installs sandboxable files too
+        self.root = root or os.path.expanduser(os.path.join("~", ".%s" % self.sprinter_namespace))
+
         self.ignore_errors = ignore_errors
+
+        # path to the directory to install global files
+        self.global_path = os.path.join(self.root, ".global")
+        self.global_config_path = os.path.join(self.global_path, "config.cfg")
+        self.global_config = global_config or load_global_config(self.global_config_path)
+        
+        self.shell_util_path = os.path.join(self.global_path, "utils.sh")
+        self.main_manifest = None
+
+        # a dictionary of the errors associated with features.
+        # The key is a tuple of feature name and formula, while the value is an instance.
+        self._error_dict = defaultdict(list)
         
     @warmup
     def install(self):
@@ -132,7 +126,7 @@ class Environment(object):
         self.phase = PHASE.INSTALL
         if not self.directory.new:
             self.logger.info("Namespace %s already exists!" % self.namespace)
-            self.source = Manifest(self.directory.manifest_path)
+            self.source = load_manifest(self.directory.manifest_path)
             return self.update()
         try:
             self.logger.info("Installing environment %s..." % self.namespace)
@@ -140,8 +134,8 @@ class Environment(object):
             self.install_sandboxes()
             self.instantiate_features()
             self._specialize()
-            for feature in self._feature_dict_order:
-                self._run_action(feature, 'sync')
+            for feature in self.features.run_order:
+                self.run_action(feature, 'sync')
             self.inject_environment_config()
             self._finalize()
         except Exception:
@@ -163,10 +157,12 @@ class Environment(object):
             self.logger.info("Updating environment %s..." % self.namespace)
             self.install_sandboxes()
             self.instantiate_features()
-            self.grab_inputs(reconfigure=reconfigure)
+            # We don't grab inputs, only on install
+            # updates inputs are grabbed on demand
+            # self.grab_inputs(reconfigure=reconfigure)
             self._specialize(reconfigure=reconfigure)
-            for feature in self._feature_dict_order:
-                self._run_action(feature, 'sync')
+            for feature in self.features.run_order:
+                self.run_action(feature, 'sync')
             self.inject_environment_config()
             self._finalize()
         except Exception:
@@ -183,8 +179,8 @@ class Environment(object):
             self.logger.info("Removing environment %s..." % self.namespace)
             self.instantiate_features()
             self._specialize()
-            for feature in self._feature_dict_order:
-                self._run_action(feature, 'sync')
+            for feature in self.features.run_order:
+                self.run_action(feature, 'sync')
             self.clear_all()
             self.directory.remove()
             self.injections.commit()
@@ -203,9 +199,9 @@ class Environment(object):
             self.directory.rewrite_config = False
             self.instantiate_features()
             self._specialize()
-            for feature in self._feature_dict_order:
+            for feature in self.features.run_order:
                 self.logger.info("Deactivating %s..." % feature[0])
-                self._run_action(feature, 'deactivate')
+                self.run_action(feature, 'deactivate')
             self.clear_all()
             self._finalize()
         except Exception:
@@ -223,9 +219,9 @@ class Environment(object):
             self.directory.rewrite_config = False
             self.instantiate_features()
             self._specialize()
-            for feature in self._feature_dict_order:
+            for feature in self.features.run_order:
                 self.logger.info("Activating %s..." % feature[0])
-                self._run_action(feature, 'activate')
+                self.run_action(feature, 'activate')
             self.inject_environment_config()
             self._finalize()
         except Exception:
@@ -244,16 +240,16 @@ class Environment(object):
             for s in self.target.formula_sections():
                 context_dict["%s:root_dir" % s] = self.directory.install_directory(s)
                 context_dict['config:root_dir'] = self.directory.root_dir
-                context_dict['config:node'] = self.system.node
+                context_dict['config:node'] = system.NODE
                 self.target.add_additional_context(context_dict)
-        for feature in self._feature_dict_order:
-            self._run_action(feature, 'validate', run_if_error=True)
+        for feature in self.features.run_order:
+            self.run_action(feature, 'validate', run_if_error=True)
 
     @warmup
     def inject_environment_config(self):
         for shell in SHELL_CONFIG:
             if shell == 'gui':
-                if self.system.isDebianBased():
+                if system.is_debian():
                     self._inject_config_source(".env", SHELL_CONFIG['gui']['debian'])
             else:
                 if (self.global_config.has_option('shell', shell)
@@ -276,7 +272,7 @@ class Environment(object):
                             source_template % (full_rc_path, full_rc_path))
                     else:
                         self.global_injections.inject(full_env_path, '')
-                    if self.system.isOSX() and not self.injections.in_noninjected_file(env_path, rc_file):
+                    if system.is_osx() and not self.injections.in_noninjected_file(env_path, rc_file):
                         if self.phase is PHASE.INSTALL:
                             self.logger.info("On OSX, login shell are the default, which only source config files")
 
@@ -289,7 +285,7 @@ class Environment(object):
 
     def install_sandboxes(self):
         if self.target:
-            if self.system.isOSX():
+            if system.is_osx():
                 if not self.target.is_affirmative('config', 'use_global_packagemanagers'):
                     self._install_sandbox('brew', brew.install_brew)
                 elif lib.which('brew') is None:
@@ -305,51 +301,55 @@ class Environment(object):
                                  output_log_level=logging.DEBUG, stdout=None)
                         brew.install_brew('/usr/local')
 
+    def instantiate_features(self):
+        if hasattr(self, 'features') and self.features:
+            return
+        self.features = FeatureDict(self,
+                                    self.source, self.target,
+                                    self.global_path)
+
     def run_feature(self, feature, action):
-        for k in self._feature_dict_order:
+        for k in self.features.run_order:
             if feature in k:
-                self._run_action(k, action, run_if_error=True)
+                self.run_action(k, action, run_if_error=True)
 
     def write_debug_log(self, file_path):
         """ Write the debug log to a file """
-        if self.write_files:
-            with open(file_path, "w+") as fh:
-                fh.write(self._debug_stream.getvalue())
-                fh.write("The following errors occured:\n")
-                for error in self._errors:
-                    fh.write(error + "\n")
-                for k, v in self._error_dict.items():
-                    if len(v) > 0:
-                        fh.write("Error(s) in %s with formula %s:\n" % k)
-                        for error in v:
-                            fh.write(error + "\n")
+        with open(file_path, "w+") as fh:
+            fh.write(self._debug_stream.getvalue())
+            fh.write("The following errors occured:\n")
+            for error in self._errors:
+                fh.write(error + "\n")
+            for k, v in self._error_dict.items():
+                if len(v) > 0:
+                    fh.write("Error(s) in %s with formula %s:\n" % k)
+                    for error in v:
+                        fh.write(error + "\n")
 
     def write_manifest(self):
         """ Write the manifest to the file """
-        if os.path.exists(self.directory.manifest_path) and self.write_files:
-            manifest = self.target or self.source
-            manifest.write(open(self.directory.manifest_path, "w+"))
+        if os.path.exists(self.directory.manifest_path):
+            self.main_manifest.write(open(self.directory.manifest_path, "w+"))
 
     def message_failure(self):
         """ return a failure message, if one exists """
-        manifest = self.target or self.source
-        if manifest and manifest.has_option('config', 'message_failure'):
-            return manifest.get('config', 'message_failure')
+        if not isinstance(self.main_manifest, Manifest):
+            return None
+        return self.main_manifest.get('config', 'message_failure', default=None)
 
     def message_success(self):
         """ return a success message, if one exists """
-        manifest = self.target or self.source
-        if manifest.has_option('config', 'message_success'):
-            return manifest.get('config', 'message_success')
+        return self.main_manifest.get('config', 'message_success', default=None)
 
     def warmup(self):
         """ initialize variables necessary to perform a sprinter action """
         self.logger.debug("Warming up...")
         try:
             if not isinstance(self.source, Manifest) and self.source:
-                self.source = Manifest(self.source)
+                self.source = load_manifest(self.source)
             if not isinstance(self.target, Manifest) and self.target:
-                self.target = Manifest(self.target)
+                self.target = load_manifest(self.target)
+            self.main_manifest = self.target or self.source
         except lib.BadCredentialsException:
             e = sys.exc_info()[1]
             self.logger.error(str(e))
@@ -374,45 +374,6 @@ class Environment(object):
         os.environ['PATH'] = self.directory.bin_path() + ":" + os.environ['PATH']
         self.warmed_up = True
 
-    def instantiate_features(self):
-        """ Create and instantiate the feature dictionary """
-        self._feature_dict = {}
-        self._feature_dict_order = []
-
-        if self.target:
-            for feature in self.target.formula_sections():
-                feature_key = self._instantiate_feature(
-                    feature, self.target.get_feature_config(feature), 'target')
-                if feature_key:
-                    self._feature_dict_order.append(feature_key)
-        if self.source:
-            for feature in self.source.formula_sections():
-                feature_key = self._instantiate_feature(
-                    feature, self.source.get_feature_config(feature), 'source')
-                if feature_key:
-                    self._feature_dict_order.insert(0, feature_key)
-
-    def _instantiate_feature(self, feature, feature_config, kind):
-        if feature_config.has('formula'):
-            key = (feature, feature_config.get('formula'))
-            if key not in self._feature_dict:
-                try:
-                    formula_class = self._get_formula_class(feature_config.get('formula'))
-                    self._feature_dict[key] = formula_class(self, feature, **{kind: feature_config})
-                    self._error_dict[key] = []
-                    if self._feature_dict[key].should_run():
-                        return key
-                    else:
-                        del(self._feature_dict[key])
-                except SprinterException:
-                    self.log_error("ERROR: Invalid formula %s for %s feature %s!"
-                                   % (feature_config.get('formula'), kind, feature))
-            else:
-                setattr(self._feature_dict[key], kind, feature_config)
-        else:
-            self.log_error('feature %s has no formula!' % feature)
-        return None
-
     def _inject_config_source(self, source_filename, files_to_inject):
         """
         Inject existing environmental config with namespace sourcing.
@@ -441,31 +402,37 @@ class Environment(object):
         """ command to run at the end of sprinter's run """
         self.logger.info("Finalizing...")
         self.write_manifest()
+
         if self.directory.rewrite_config:
             # always ensure .rc is written (sourcing .env)
             self.directory.add_to_rc('')
             # prepend brew for global installs
-            manifest = self.target or self.source
-            if self.system.isOSX() and manifest.is_affirmative('config', 'use_global_packagemanagers'):
-                self.directory.add_to_env('sprinter_prepend_path "%s" PATH' % '/usr/local/bin')
-            self.directory.add_to_env('sprinter_prepend_path "%s" PATH' % self.directory.bin_path())
-            self.directory.add_to_env('sprinter_prepend_path "%s" LIBRARY_PATH' % self.directory.lib_path())
-            self.directory.add_to_env('sprinter_prepend_path "%s" C_INCLUDE_PATH' % self.directory.include_path())
-        if self.write_files:
-            self.injections.commit()
-            self.global_injections.commit()
-            if not os.path.exists(os.path.join(self.root, ".global")):
-                self.logger.debug("Global directoy doesn't exist! creating...")
-                os.makedirs(os.path.join(self.root, ".global"))
-            self.logger.debug("Writing global config...")
-            self.global_config.write(open(os.path.join(self.root, ".global", "config.cfg"), 'w+'))
-            self.logger.debug("Writing shell util file...")
-            with open(self.shell_util_path, 'w+') as fh:
-                fh.write(shell_utils_template)
+            if system.is_osx() and self.main_manifest.is_affirmative('config', 'use_global_packagemanagers'):
+                self.directory.add_to_env('__sprinter_prepend_path "%s" PATH' % '/usr/local/bin')
+            self.directory.add_to_env('__sprinter_prepend_path "%s" PATH' % self.directory.bin_path())
+            self.directory.add_to_env('__sprinter_prepend_path "%s" LIBRARY_PATH' % self.directory.lib_path())
+            self.directory.add_to_env('__sprinter_prepend_path "%s" C_INCLUDE_PATH' % self.directory.include_path())
+
+        self.injections.commit()
+        self.global_injections.commit()
+
+        if not os.path.exists(os.path.join(self.root, ".global")):
+            self.logger.debug("Global directory doesn't exist! creating...")
+            os.makedirs(os.path.join(self.root, ".global"))
+
+        self.logger.debug("Writing global config...")
+        self.global_config.write(open(self.global_config_path, 'w+'))
+
+        self.logger.debug("Writing shell util file...")
+        with open(self.shell_util_path, 'w+') as fh:
+            fh.write(shell_utils_template)
+
         if self.error_occured:
             raise SprinterException("Error occured!")
+
         if self.message_success():
             self.logger.info(self.message_success())
+
         self.logger.info("NOTE: Please remember to open new shells/terminals to use the modified environment")
 
     def _install_sandbox(self, name, call, kwargs={}):
@@ -490,44 +457,26 @@ class Environment(object):
         logger.setLevel(logging.DEBUG)
         return logger
 
-    def _get_formula_class(self, formula):
-        """
-        get a formula class object if it exists, else
-        create one, add it to the dict, and pass return it.
-        """
-        formula_class, formula_url = formula, None
-        if ':' in formula:
-            formula_class, formula_url = formula.split(":", 1)
-        if formula_class not in self.formula_dict:
-            try:
-                self.formula_dict[formula_class] = lib.get_subclass_from_module(formula_class, FormulaBase)
-            except (SprinterException, ImportError):
-                self.logger.info("Downloading %s..." % formula_class)
-                try:
-                    self._pip.install_egg(formula_url or formula_class)
-                    try:
-                        self.formula_dict[formula_class] = lib.get_subclass_from_module(formula_class, FormulaBase)
-                    except ImportError:
-                        raise SprinterException("Error: Unable to retrieve formula %s!" % formula_class)
-                except PipException:
-                    self.logger.error("ERROR: Unable to download %s!" % formula_class)
-        return self.formula_dict[formula_class]
-
     def log_error(self, error_message):
         self.error_occured = True
         self._errors += [error_message]
         self.logger.error(error_message)
 
+    def get_error_value(self, feature):
+        """ get the error value for a feature """
+        if feature not in self._error_dict:
+            self._error_dict1
+
     def log_feature_error(self, feature, error_message):
         self.error_occured = True
         self._error_dict[feature] += [error_message]
         self.logger.error(error_message)
             
-    def _run_action(self, feature, action, run_if_error=False):
+    def run_action(self, feature, action, run_if_error=False):
         """ Run an action, and log it's output in case of errors """
         if len(self._error_dict[feature]) > 0 and not run_if_error:
             return
-        instance = self._feature_dict[feature]
+        instance = self.features[feature]
         try:
             result = getattr(instance, action)()
             if result:
@@ -555,14 +504,14 @@ class Environment(object):
                 for s in manifest.formula_sections():
                     context_dict["%s:root_dir" % s] = self.directory.install_directory(s)
                     context_dict['config:root_dir'] = self.directory.root_dir
-                    context_dict['config:node'] = self.system.node
+                    context_dict['config:node'] = system.NODE
                 manifest.add_additional_context(context_dict)
         self.grab_inputs()
-        for feature in self._feature_dict_order:
-            self._run_action(feature, 'validate', run_if_error=True)
+        for feature in self.features.run_order:
+            self.run_action(feature, 'validate', run_if_error=True)
             if not reconfigure:
-                self._run_action(feature, 'resolve')
-            self._run_action(feature, 'prompt')
+                self.run_action(feature, 'resolve')
+            self.run_action(feature, 'prompt')
 
     def grab_inputs(self, reconfigure=False):
         """ Resolve the source and target config section """
@@ -572,62 +521,4 @@ class Environment(object):
                     if not self.target.has_option('config', k):
                         self.target.set('config', k, v)
         if self.target:
-            self.target.grab_inputs(force_prompt=reconfigure)
-
-    def load_global_config(self, global_config_string):
-        if self.global_config:
-            return self.global_config
-        self.global_config = configparser.RawConfigParser()
-        if global_config_string:
-            self.global_config.readfp(StringIO(global_config_string))
-        else:
-            global_config_path = os.path.join(self.root, ".global", "config.cfg")
-            if os.path.exists(global_config_path):
-                self.global_config.read(global_config_path)
-                self.logger.info("Checking and setting global parameters...")
-            else:
-                self.initial_run()  # it's probably the first run, so stick initial run logic in here
-                self.logger.info("Unable to find a global sprinter configuration!")
-                self.logger.info("Creating one now. Please answer some questions" +
-                                 " about what you would like sprinter to do.")
-                self.logger.info("")
-            if not self.global_config.has_section('shell'):
-                # shell configuration
-                self.global_config.add_section('shell')
-                self.logger.info("What shells or environments would you like sprinter to work with?\n" +
-                                 "(Sprinter will not try to inject into environments not specified here.)\n" +
-                                 "If you specify 'gui', sprinter will attempt to inject it's state into graphical programs as well.\n" +
-                                 "i.e. environment variables sprinter set will affect programs as well, not just shells")
-                environments = list(enumerate(sorted(SHELL_CONFIG), start=1))
-                self.logger.info("[0]: All, " + ", ".join(["[%d]: %s" % (index, val) for index, val in environments]))
-                desired_environments = lib.prompt("type the environment, comma-separated", default="0")
-                for index, val in environments:
-                    if str(index) in desired_environments or "0" in desired_environments:
-                        self.global_config.set('shell', val, 'true')
-                    else:
-                        self.global_config.set('shell', val, 'false')
-            if not self.global_config.has_section('global'):
-                self.global_config.add_section('global')
-            if not self.global_config.has_option('global', 'env_source_rc'):
-                self.global_config.set('global', 'env_source_rc', False)
-                if self.system.isOSX():
-                    self.logger.info("On OSX, login shells are default, which only source sprinter's 'env' configuration.")
-                    self.logger.info("I.E. environment variables would be sourced, but not shell functions "
-                                     + "or terminal status lines.")
-                    self.logger.info("The typical solution to get around this is to source your rc file (.bashrc, .zshrc) "
-                                     + "from your login shell.")
-                    env_source_rc = lib.prompt("would you like sprinter to source the rc file too?", default="yes",
-                                               boolean=True)
-                    self.global_config.set('global', 'env_source_rc', env_source_rc)
-
-    def initial_run(self):
-        """ A method that only runs during the initial run of sprinter """
-        if not self.system.is_officially_supported():
-            self.logger.warn(warning_template
-                             + "===========================================================\n"
-                             + "Sprinter is not officially supported on {0}! Please use at your own risk.\n\n".format(self.system.operating_system())
-                             + "You can find the supported platforms here:\n"
-                             + "(http://sprinter.readthedocs.org/en/latest/index.html#compatible-systems)\n\n"
-                             + "Conversely, please help us support your system by reporting on issues\n"
-                             + "(http://sprinter.readthedocs.org/en/latest/faq.html#i-need-help-who-do-i-talk-to)\n"
-                             + "===========================================================")
+            self.target.grab_inputs(force=reconfigure)
